# Generated by Django makemigrations for adding status field to Club model

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0051_plpgsql_vitriolic_stage_group_position"),
    ]

    operations = [
        migrations.AddField(
            model_name="club",
            name="status",
            field=models.CharField(
                choices=[
                    ("active", "Active"),
                    ("inactive", "Inactive"),
                    ("hidden", "Hidden"),
                ],
                db_index=True,
                default="active",
                help_text="Current status of the club.",
                max_length=20,
            ),
        ),
    ]